CON
'' ******************************************************************************
'' * Compass test program.
'' *
'' * Reads compass heading from a parallax HM55B and then causes one of 16
'' * LEDs to light up to indicate North. 16 LEDs use a lot of I/O pins, so
'' * I use an I2C port expander (MCP23017) to drive them using only 2 Propeller
'' * Pins.
'' *
'' * The MCP23017 is based on:
'' *    I2C Demo Propeller program
'' *    James Burrows Oct 2007
'' *    Version 2.1
'' *
'' * Uses :  Mike Green's "basic_i2c_driver" with a couple of extra bits
'' *
'' ******************************************************************************
''
''


CON
  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000
  _stack        = 50

  ioPin        = 28

  ioDeviceAddress  = %0100_1000
 ' debug - USE onboard pins
  pcDebugRX       = 31
  pcDebugTX       = 30
  ' serial baud rate
  pcDebugBaud     = 115200

  on     =  1
  off    =  0
  toggle = -1

  Enable = 27
  Clock  = 26
  Data   = 25



' MCP23017 Pin addresses
  GPA0 = 00
  GPA1 = 01
  GPA2 = 02
  GPA3 = 03
  GPA4 = 04
  GPA5 = 05       'LED
  GPA6 = 06       'LED
  GPA7 = 07       'LED

  GPB0 = 10       'LED
  GPB1 = 11       'LED
  GPB2 = 12       'LED
  GPB3 = 13       'LED
  GPB4 = 14       'LED
  GPB5 = 15
  GPB6 = 16
  GPB7 = 17

  ledNorth  = GPB5
  ledNNE    = GPB4
  ledNEast  = GPB3
  ledENE    = GPB2

  ledEast   = GPB1
  ledESE    = GPB0
  ledSEast  = GPA7
  ledSSE    = GPA6

  ledSouth  = GPA5
  ledSSW    = GPA4
  ledSWest  = GPA3
  ledWSW    = GPA2

  ledWest   = GPA1
  ledWNW    = GPA0
  ledNWest  = GPB7
  ledNNW    = GPB6

CON     ''Setup Constants for the Compass Calibration
        Samples = 25                                    'Number of samples to average



VAR
  long  i2cAddress, i2cSlaveCounter
  long  leds[16]
  byte  port[2]

VAR     ''Setup variables related to the compass
    long  Deg, Index, DegDataBase, DegAverage
    long  HeadingIndex, CalibrationDelta, CorrectHeading, HeadingError
    long  LowCalibration, HighCalibration

VAR     ''Setup variables related to converting numbers to strings for display
    long  idx           ' ( 1 long ) pointer into string
    byte  nstr[64]      ' (16 longs) string
    byte  z_pad


OBJ
  IO             : "cr_MCP23017Object"

  'debug          : "Debug_PC"
  bit            : "obex_BitManipulator"
  HM55B          : "HM55B Compass Module Asm"


pub Start
    'debug.startx(pcDebugRX,pcDebugTX,pcDebugBaud)
    'debug.putc(13)
    'debug.putc(13)
    'debug.putc(13)
    'debug.str(string("Starting..."))

    'start and setup Compass
    waitcnt(clkfreq/4 + cnt)
    HM55B.start(Enable,Clock,Data)                      '' Initialize Compass Object
    'debug.putc(".")

    waitcnt(clkfreq/4 + cnt)

    IO.Initialize(ioPin)

    leds[15] := ledNNE
    leds[14] := ledNEast
    leds[13] := ledENE
    leds[12] := ledEast
    leds[11] := ledESE
    leds[10] := ledSEast
    leds[ 9] := ledSSE
    leds[ 8] := ledSouth
    leds[ 7] := ledSSW
    leds[ 6] := ledSWest
    leds[ 5] := ledWSW
    leds[ 4] := ledWest
    leds[ 3] := ledWNW
    leds[ 2] := ledNWest
    leds[ 1] := ledNNW
    leds[ 0] := ledNorth

    IO.setRegister(ioDeviceAddress,IO#cIODIRA,%0000_0000)
    waitcnt(clkfreq/4 + cnt)
    IO.setRegister(ioDeviceAddress,IO#cIODIRB,%0000_0000)
    waitcnt(clkfreq/4 + cnt)

    IO.setRegister(ioDeviceAddress,IO#cOLATA,%1111_1111)
    IO.setRegister(ioDeviceAddress,IO#cOLATB,%1111_1111)
    waitcnt(clkfreq*2 + cnt)

    IO.setRegister(ioDeviceAddress,IO#cOLATA,%0000_0000)
    IO.setRegister(ioDeviceAddress,IO#cOLATB,%0000_0000)

    'i2cScan
    MCP23017_Test
    'Calibration                                         '' Start Calibration Sequence
    'debug.putc(".")
    Compass_Demo                                        '' Start the Compass DEMO
    'debug.putc(".")


    repeat

        if IO.devicePresent(ioDeviceAddress) == true
            MCP23017_Test

PUB DetectRange(RAW_Deg)|IndexHead
    repeat IndexHead from 0 to 15
      if RAW_Deg < CompassCalibration[IndexHead]         ' Determine position from Calibration values
         If IndexHead <> 0
            LowCalibration := CompassCalibration[IndexHead-1]
            HighCalibration := CompassCalibration[IndexHead]
            HeadingIndex := (IndexHead-1) * 512
            quit
      else
         LowCalibration := CompassCalibration[15]        ' Compensate for 360 / 0 Deg transition
         HighCalibration := CompassCalibration[0] + 8192
         HeadingIndex := 7680



PUB Compass_Demo | RawHeading,i,n,cntr

    'debug.putc(13)

    repeat
      if cntr++ >1000      ' Do this "Every so often"
         cntr:=0
         MCP23017_Test     ' Cause the test routine to rotate through all the LEDs - Makes a nice light show, but not needed.

      Deg := getRawHeading
      n:= (((10*deg)-112)/225)

      'debug.putc(13)
      'debug.decf( Deg,3 )
      'debug.putc("-")
      'debug.decf( n,3 )

      repeat i from 0 to 15
        case (n-i)
           0:
              ledOn(i)
              'debug.decf( i,3 )
              'debug.putc(" ")
           other :
              ledOff(i)


PUB getRawHeading | RawHeading,i,n

      RawHeading := HM55B.Theta                         ' Read RAW 13-bit Angle

      If RawHeading < 0                                 ' Compensate for the 360 Deg / 0 Deg transition
         RawHeading := RawHeading + 8192

      Deg := RawHeading * 45 / 1024                    ' Convert 13-Bit Angle to Deg
      Return Deg

PUB getCorrectedHeading | RawHeading,i,n

      RawHeading := HM55B.Theta                         ' Read RAW 13-bit Angle

      DetectRange(RawHeading)                           ' Based on calibrated measurements detect
                                                        ' which 1/16th section of the compass circle
                                                        ' the compass is facing
      CalibrationDelta := HighCalibration - LowCalibration  ' Determine amount of difference in the 1/16 compass section.

      HeadingError := (512 * 1000) / CalibrationDelta   ' Determine the amount of error or
                                                        ' scaling that needs to be done based on how
                                                        ' much should be within 1/16th of the compass
                                                        ' section vs. how much is actually within 1/16th
                                                        ' of the compass area.
                                                        '
                                                        ' Note: With a 13-bit returned compass angle,
                                                        '       1/16th of the total compass circle equates
                                                        '       to 512 points.

      CorrectHeading := (((RawHeading - LowCalibration) * HeadingError) / 1000)+ HeadingIndex     'Calculate
                                                        ' correct heading by creating a linear relationship
                                                        ' between calibrated values vs. actual values

      If CorrectHeading < 0                             ' Compensate for the 360 Deg / 0 Deg transition
         CorrectHeading := CorrectHeading + 8192


      Deg := CorrectHeading * 45 / 1024                 ' Convert 13-Bit Angle to Deg

      Return Deg


PUB Calibration | Orientation
    'debug.strln(string("Calibration "))

    repeat

      repeat index from 0 to 15
        if CompassCalibration[index] == 8192             ' Test for default calibration value
           'debug.putc(13)
           'debug.dec( index)
           'debug.putc("-")
           'debug.str(string("Point the compass "))

           Orientation := decx(index*225,4)              ' Determine missing calibration data
           nstr[4] := nstr[3]                            ' Trick to insert ...
           nstr[3] := 46                                 ' ... a decimal point

           repeat until nstr[0] <> 48                    ' Left justify Leading Zero's
             bytemove (@nstr[0],@nstr[1],5)

           if index <> 0

             'debug.dec( Orientation)
           else
             'debug.str(string("0.0"))

           'debug.str(string(" and update result in the Compass Calibration DAT section of this program with the value displayed below."))
           'debug.str(string(" and set to "))

           Deg := HM55B.Theta                            ' Read RAW 13-bit Angle
           Deg := Deg / 11                               ' Remove noise down to about 1/2 a deg
           DegDataBase := DegDataBase - DegAverage + Deg ' Calculate average
           DegAverage := DegDataBase / Samples
           Deg := DegAverage * 11                        ' Restore to 13-bit Angle format

           'debug.dec(Deg)                  ' Display Average Deg
           'debug.putc(" ")
           'waitcnt(clkfreq/4 + cnt)
           QUIT
        if index == 15                                   ' If index = 15 then no calibration
           RETURN



PRI decstr(value) | div

' Converts value to signed-decimal string equivalent
' -- characters written to current position of idx
' -- returns pointer to nstr

  if (value < 0)                                         ' negative value?
    -value                                               '   yes, make positive
    nstr[idx++] := "-"                                   '   and print sign indicator

  div := 1_000_000_000                                   ' initialize divisor
  z_pad~                                                 ' clear zero-pad flag

  repeat 10
    if (value => div)                                    ' printable character?
      nstr[idx++] := (value / div + "0")                 '   yes, print ASCII digit
      value //= div                                      '   update value
      z_pad~~                                            '   set zflag
    elseif z_pad or (div == 1)                           ' printing or last column?
      nstr[idx++] := "0"
    div /= 10

  return @nstr


PRI Dec(value)
    bytefill(@nstr, 0, 64)                               ' clear string to zeros
    idx~                                                 ' reset index
    return decstr(value)

PUB Decx(value, digits) | div

'' Returns pointer to zero-padded, signed-decimal string
'' -- if value is negative, field width is digits+1

    bytefill(@nstr, 0, 64)                               ' clear string to zeros
    idx~                                                 ' reset index
    digits := 1 #> digits <# 10

    if (value < 0)                                       ' negative value?
      -value                                             '   yes, make positive
      nstr[idx++] := "-"                                 '   and print sign indicator
      div := 1_000_000_000                               ' initialize divisor
    if digits < 10                                       ' less than 10 digits?
      repeat (10 - digits)                               '   yes, adjust divisor
        div /= 10

    value //= (div * 10)                                 ' truncate unused digits

    repeat digits
      nstr[idx++] := (value / div + "0")                 ' convert digit to ASCII
      value //= div                                      ' update value
      div /= 10                                          ' update divisor

    return @nstr



PRI MCP23017_Test | counter

  repeat 1
    repeat counter from 0 to 15
        ledOn(counter)
        waitcnt(clkfreq/25 + cnt)

    repeat counter from 15 to 0
        ledOff(counter)
        waitcnt(clkfreq/25 + cnt)

    repeat counter from 15 to 0
        ledOn(counter)
        waitcnt(clkfreq/25 + cnt)

    repeat counter from 0 to 15
        ledOff(counter)
        waitcnt(clkfreq/25 + cnt)




pub allOn
    IO.setRegister(ioDeviceAddress,IO#cOLATA,-1)
    IO.setRegister(ioDeviceAddress,IO#cOLATB,-1)

pub allOff
    IO.setRegister(ioDeviceAddress,IO#cOLATA,0)
    IO.setRegister(ioDeviceAddress,IO#cOLATB,0)

pub ledOn(bitNumber) | n,p0,p1
        p0 := 0
        p1 := 0

        n := leds[bitNumber]
        if n<10
          p0 := n
          port[0] := port[0] | (|<p0)
          IO.setRegister(ioDeviceAddress,IO#cOLATA,port[0])
        else
          p1 := n // 10
          port[1] := port[1] | (|<p1)
          IO.setRegister(ioDeviceAddress,IO#cOLATB,port[1])


pub ledOff(bitNumber) | n,p0,p1
        p0 := 0
        p1 := 0

        n := leds[bitNumber]
        if n<10
          p0 := n
          port[0] := port[0] & (255-(|<p0))
          IO.setRegister(ioDeviceAddress,IO#cOLATA,port[0])
        else
          p1 := n // 10
          port[1] := port[1] & (255-(|<p1))
          IO.setRegister(ioDeviceAddress,IO#cOLATB,port[1])


PRI i2cScan | value, ackbit
    ' Scan the I2C Bus and debug the LCD
    'debug.strln(string("Scanning I2C Bus...."))

    ' initialize variables
    i2cSlaveCounter := 0

    ' i2c Scan - scans all the address's on the bus
    ' sends the address byte and listens for the device to ACK (hold the SDA low)

    repeat i2cAddress from 0 to 127

        value :=  i2cAddress << 1 | 0
        ackbit := IO.devicePresent(value)

        if ackbit==false
            'debug.str(string("NAK"))
        else
            ' show the scan
            'debug.str(string("Scan Addr :  %"))
            'debug.bin(value,8)
            'debug.str(string(",  ACK"))

            ' the device has set the ACK bit
            i2cSlaveCounter ++
            waitcnt(clkfreq/10+cnt)

            'debug.putc(13)

        ' slow the scan so we can read it.
        waitcnt(clkfreq/400 + cnt)

    ' update the counter
    'debug.str(string("i2cScan found "))
    'debug.dec(i2cSlaveCounter)
    'debug.strln(string(" devices!"))
    waitcnt(clkfreq + cnt)


DAT     ''Data setup section for compass calibration values

''Compass Calibration section requires 8 longs of storage

''                         Raw Reading     Actual Direction

'                               8192                     ' <--Default Calibration value for all Deg.
'                                                          Since Deg only ranges from 0 to 8191, a
'                                                          value of 8192 is used by the calibration
'                                                          routine to determine if calibration for
'                                                          that position is necessary.
'
'                         My Compass values
CompassCalibration      word    304                     '   0.0 Deg     <---- TIP: To Calibrate Compass, just set this
                        word    1056                     '  22.5 Deg                value to 8192 and run the program.
                        word    1595                     '  45.0 Deg                Move the compass orientation to each
                        word    1991                     '  67.5 Deg                22.5 deg position and write down or
                        word    2376                     '  90.0 Deg                enter the values in the table to the
                        word    2783                     ' 112.5 Deg                left.  Once the table is complete, the
                        word    3190                     ' 135.0 Deg                the compass is now calibrated.  Re-load
                        word    3652                     ' 157.5 Deg                and run the program.  The compass display
                        word    4037                     ' 180.0 Deg                will show the RAW heading in RED, and the
                        word    4345                     ' 202.5 Deg                correct calibrated heading in YELLOW.
                        word    4763                     ' 225.0 Deg
                        word    5214                     ' 247.5 Deg                Refer to page 3 of the document link below
                        word    5687                     ' 270.0 Deg                for easy compass orientation and calibration.
                        word    6215                     ' 292.5 Deg
                        word    7040                     ' 315.0 Deg
                        word    7821                     ' 337.5 Deg

'                                              http://www.parallax.com/Portals/0/Downloads/docs/prod/compshop/HM55BModDocs.pdf

