'' ******************************************************************************
'' * MCP320017 i2c Bus Expander Object                                          *
'' * Written Jan 2009 by Chuck Rice                                             *
'' * Based on the OBEX MCP320016 version by James Burrows Oct 07                *
'' * Version 0.0                                                                *
'' ******************************************************************************
''
'' MCP23017 i2c I/O expander object.
''
'' For a simple example write a 0 to the I/O register to set all 8 pins output, and
'' 255 or %1111_1111 to GP0 or GP1 to make the pins +5V.  Attach some LED's!
''
'' www.microchip.com
''
'' this object provides the PUBLIC functions:
''  -> WriteIOregister0 - set the output register bank 0
''  -> WriteIOregister1 - set the output register bank 1
''  -> WriteGP0 - set the GP register bank 0
''  -> WriteGP1 - set the GP register bank 1
''  -> ReadGP0 - read the GP register bank 0
''  -> ReadGP1 - read the GP register bank 0
''
'' this object provides the PRIVATE functions:
''  -> None
''
'' this object uses the following sub OBJECTS:
''  -> i2cObject
''
'' Revision History:
''  -> V2    - re-Release
''
'' The default address is %0100_0000

CON

   cIODIRA   = $00
   cIPOLA    = $01
   cGPINTENA = $02
   cDEFVALA  = $03
   cINTCONA  = $04
   cIOCONa   = $05
   cGPPUA    = $06
   cINTFA    = $07
   cINTCAPA  = $08
   cGPIOA    = $09
   cOLATA    = $0A

   cIODIRB   = $10
   cIPOLB    = $11
   cGPINTENB = $12
   cDEFVALB  = $13
   cINTCONB  = $14
   cIOCONb   = $15
   cGPPUB    = $16
   cINTFB    = $17
   cINTCAPB  = $18
   cGPIOB    = $19
   cOLATB    = $1A

OBJ
  i2cObject         : "obex_Basic_I2C_Driver"

VAR

  byte              i2cPin

PUB Initialize(i2cSCLpin)
    i2cPin := i2cSCLpin[0]
    i2cObject.Initialize(i2cPin)


PUB devicePresent(MCP23017Addr)
    return i2cObject.devicePresent(i2cPin,MCP23017Addr)


PUB setRegister(i2cDeviceAddress,i2cRegister,i2cData) : ackbit
  ' write to the I/O register
  ackbit := i2cObject.WriteLocation(i2cPin,i2cDeviceAddress, i2cRegister,i2cData)
  return ackbit


PUB getRegister(i2cDeviceAddress,i2cRegister,i2cDat) : i2cData
  ' Read the General Purpose (GP) Port register 0
  i2cData := i2cObject.readLocation(i2cPin,i2cDeviceAddress, i2cRegister)
  return i2cData



